
#include "AE_Radial_Deg_4_LD.h"
#include <ldpk/tde4_ldp_radial_deg_8.h>
#include <ldpk/tde4_ldp_radial_decentered_deg_4_cylindric.h>
#include <ldpk/tde4_ldp_classic_3de_mixed.h>
#include <ldpk/tde4_ldp_anamorphic_deg_6.h>


#include <ldpk/ldpk_vec2d.h>

static SPBasicSuite *S_pica_suiteP = NULL;

static PF_Err 
About (	
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output )
{
	AEGP_SuiteHandler suites(in_data->pica_basicP);
	
	suites.ANSICallbacksSuite1()->sprintf(	out_data->return_msg, 
											"%s, v%d.%d\r%s",
											NAME, 
											MAJOR_VERSION, 
											MINOR_VERSION, 
											DESCRIPTION);

	return PF_Err_NONE;
}


static PF_Err 
GlobalSetup (PF_InData		*in_data,
			PF_OutData		*out_data,
			PF_ParamDef		*params[],
			PF_LayerDef		*output )
{
	out_data->my_version	= PF_VERSION(	MAJOR_VERSION, 
											MINOR_VERSION,
											BUG_VERSION, 
											STAGE_VERSION, 
											BUILD_VERSION);

	out_data->out_flags = 0;

	S_pica_suiteP 	= in_data->pica_basicP;
	
	out_data->out_flags2	|=	PF_OutFlag2_SUPPORTS_SMART_RENDER	|
								PF_OutFlag2_FLOAT_COLOR_AWARE;
	return PF_Err_NONE;
}

static 
PF_Err 
FrameSetup (
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output )
{
	// Output buffer resizing may only occur during PF_Cmd_FRAME_SETUP. 

	PF_FpLong		border_x	=	0, 
					border_y	=	0,	
					border 		=	2.0;
	
	if (2.0) {
		// shrink the border to accomodate decreased resolutions.
		 
		border_x = border * 1;//(static_cast<PF_FpLong>(in_data->downsample_x.num) / in_data->downsample_x.den);
		border_y = border * 1;//(static_cast<PF_FpLong>(in_data->downsample_y.num) / in_data->downsample_y.den);
	} else 	{
		border_x = border_y = border;
	}
	
	// add 2 times the border width and height to the input width and
	// height to get the output size.
	 
	out_data->width  = 2 * static_cast<A_long>(border_x) + params[0]->u.ld.width;
	out_data->height = 2 * static_cast<A_long>(border_y) + params[0]->u.ld.height;
	
	printf("%i %i\n", out_data->width, out_data->height);
	
	// The origin of the input buffer corresponds to the (border_x, 
	// border_y) pixel in the output buffer.
	 
	out_data->origin.h = static_cast<A_short>(border_x);
	out_data->origin.v = static_cast<A_short>(border_y);

	return PF_Err_NONE;
}



static PF_Err 
ParamsSetup (	
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output)
{
	PF_Err	err = PF_Err_NONE;
	
	PF_ParamDef		def;
	AEFX_CLR_STRUCT(def);	
	PF_ADD_FIXED(	("Focal Length mm"), 
				 0, 
				 1000, 
				 0, 
				 1000, 
				 33, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 FOCAL_LENGTH_ID);
	
	
	AEFX_CLR_STRUCT(def);	
	PF_ADD_FIXED(	("Filmback Width cm"), 
				 0, 
				 500, 
				 0, 
				 500, 
				 3.14, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 FB_WIDTH_ID);
	AEFX_CLR_STRUCT(def);	
	PF_ADD_FIXED(	("Filmback Height cm"), 
				 0, 
				 500, 
				 0, 
				 500, 
				 3.14, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 FB_HEIGHT_ID);
				 
	PF_ADD_POINT(	"FOV Min", 
					0, 
					0, 
					RESTRICT_BOUNDS,
					FOV_MIN_ID);			 
	PF_ADD_POINT(	"FOV Max", 
					100, 
					100, 
					RESTRICT_BOUNDS,
					FOV_MAX_ID);			 
				 
	AEFX_CLR_STRUCT(def);
	PF_ADD_FIXED(	("Lens Center Offset x cm"), 
				 -500, 
				 500, 
				 -500, 
				 500, 
				 0, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 LENS_CENTER_OFFSET_X_ID);	
	AEFX_CLR_STRUCT(def);
	PF_ADD_FIXED(	("Lens Center Offset y cm"), 
				 -500, 
				 500, 
				 -500, 
				 500, 
				 0, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 LENS_CENTER_OFFSET_Y_ID);	
	AEFX_CLR_STRUCT(def);
	PF_ADD_FIXED(	("Pixel Aspect"), 
				 0, 
				 10, 
				 0, 
				 10, 
				 1, 
				 SLIDER_PRECISION, 
				 0,
				 0,
				 PIXEL_ASPECT_ID);
	
	AEFX_CLR_STRUCT(def);
	def.flags		= PF_ParamFlag_CANNOT_TIME_VARY	| PF_ParamFlag_CANNOT_INTERP;
	PF_ADD_POPUP("Distortion Mode", 
				 2,
				 2,
				 "Apply Distortion|Remove Distortion",
				 DISTORT_MODE_ID);

				 
	tde4_ld_plugin *plug;
	
	plug = (tde4_ld_plugin*) new tde4_ldp_radial_decentered_deg_4_cylindric<ldpk::vec2d,ldpk::mat2d>();
	
	int id_counter = DISTORT_MODE_ID + 1;
	
	
	
	int nParams = 0;
	
	plug->getNumParameters(nParams);
	
	for(int i =0;i < nParams;i++)
	{
		AEFX_CLR_STRUCT(def);
		char paramName[1000];
		plug->getParameterName(i, paramName);
		tde4_ldp_ptype type;
		plug->getParameterType(paramName, type);
		switch (type) {
			case TDE4_LDP_STRING:
				break;
			case TDE4_LDP_DOUBLE:
			{
				double pDefault;
				plug->getParameterDefaultValue(paramName,pDefault);
				PF_ADD_FIXED(paramName, 
							 -10, 
							 10, 
							 0, 
							 10, 
							 pDefault, 
							 SLIDER_PRECISION, 
							 0,
							 0,
							 id_counter++);				
			}
				break;
			case TDE4_LDP_INT:
			{
				int pDefault;
				plug->getParameterDefaultValue(paramName,pDefault);
				PF_ADD_FIXED(paramName, 
							 -10, 
							 10, 
							 0, 
							 10, 
							 pDefault, 
							 0, 
							 0,
							 0,
							 id_counter++);	
			}
			case TDE4_LDP_FILE:
				break;
			case TDE4_LDP_TOGGLE:
				break;
			case TDE4_LDP_ADJUSTABLE_DOUBLE:
			{
				double pDefault;
				plug->getParameterDefaultValue(paramName,pDefault);
				double dmin,dmax;
				plug->getParameterRange(paramName, dmin,dmax);
				double c = (dmax + dmin) * 0.5;
				dmin = c - std::abs(c-dmin) * 3.0;
				dmax = c + std::abs(dmax-c) * 3.0;
				PF_ADD_FIXED(paramName, 
							 dmin, 
							 dmax, 
							 0, 
							 10, 
							 pDefault, 
							 SLIDER_PRECISION, 
							 0,
							 0,
							 id_counter++);	
			}
				break;
				
			default:
				break;
		}
		
	}
	
	
	out_data->num_params = LDP_NUM_PARAMS+nParams; 
	
	delete plug;
	
	return err;
}

static PF_Err 
DistortImage8 (
	void* 		refcon, 
	A_long 		xL, 
	A_long 		yL, 
	PF_Pixel 	*inP, 
	PF_Pixel 	*outP)
{
	register DistInfo	*siP		= reinterpret_cast<DistInfo*>(refcon);
	PF_Err				err			= PF_Err_NONE;
	PF_Fixed			new_xFi		= 0, 
						new_yFi		= 0;
						
	if (S_pica_suiteP){
		AEGP_SuiteHandler suites(S_pica_suiteP);

		A_long cor_x = xL;
		A_long cor_y = siP->footage_height - yL;
		
		double fov_min_x = siP->_fov[0];
		double fov_max_x = siP->_fov[2];
		double fov_min_y = siP->_fov[1];
		double fov_max_y = siP->_fov[3];
		
		
		double x0,x1,y0,y1;
		
		x1 = x0 = (double(cor_x)+0.5 - fov_min_x) / double(fov_max_x - fov_min_x);
		y1 = y0 = (double(cor_y)+0.5 - fov_min_y) / double(fov_max_y - fov_min_y);
		
		
		if(siP->plugin)
		{
			switch (siP->distortion_mode) 
			{
				case DM_DISTORT:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
				case DM_UNDISTORT:
					siP->plugin->distort(x0, y0, x1, y1);
					break;
				default:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
			}
		}
		double xt = x1;
		double yt = y1;
		
		new_xFi = FLOAT2FIX(double(fov_max_x - fov_min_x)*xt - 0.5 + fov_min_x);
		new_yFi = FLOAT2FIX(double(siP->footage_height) - (double(fov_max_y - fov_min_y)*yt - 0.5 + fov_min_y));
		
		ERR(suites.Sampling8Suite1()->subpixel_sample(	siP->in_data.effect_ref,
														new_xFi, 
														new_yFi, 
														&siP->samp_pb, 
														outP));
	} else {
		err = PF_Err_INTERNAL_STRUCT_DAMAGED;
	}
	return err;
}
 
static PF_Err 
DistortImage16 (
	void* 		refcon, 
	A_long 		xL, 
	A_long 		yL, 
	PF_Pixel16 	*inP, 
	PF_Pixel16 	*outP)
{
	register DistInfo	*siP		= (DistInfo*)refcon;
	PF_InData			*in_data	= &(siP->in_data);
	PF_Err				err			= PF_Err_NONE;
	PF_Fixed			new_xFi 		= 0, 
						new_yFi 		= 0;
						
	AEGP_SuiteHandler suites(in_data->pica_basicP);

	A_long cor_x = xL;
		A_long cor_y = siP->footage_height - yL;
		
		double fov_min_x = siP->_fov[0];
		double fov_max_x = siP->_fov[2];
		double fov_min_y = siP->_fov[1];
		double fov_max_y = siP->_fov[3];
		
		
		double x0,x1,y0,y1;
		
		x1 = x0 = (double(cor_x)+0.5 - fov_min_x) / double(fov_max_x - fov_min_x);
		y1 = y0 = (double(cor_y)+0.5 - fov_min_y) / double(fov_max_y - fov_min_y);
		
		
		if(siP->plugin)
		{
			switch (siP->distortion_mode) 
			{
				case DM_DISTORT:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
				case DM_UNDISTORT:
					siP->plugin->distort(x0, y0, x1, y1);
					break;
				default:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
			}
		}
		double xt = x1;
		double yt = y1;
		
		new_xFi = FLOAT2FIX(double(fov_max_x - fov_min_x)*xt - 0.5 + fov_min_x);
		new_yFi = FLOAT2FIX(double(siP->footage_height) - (double(fov_max_y - fov_min_y)*yt - 0.5 + fov_min_y));
		
	ERR(suites.Sampling16Suite1()->subpixel_sample16(	in_data->effect_ref,
														new_xFi, 
														new_yFi, 
														&siP->samp_pb, 
														outP));
	return err;
}

static PF_Err 
DistortImageFloat (
	void* 			refcon, 
	A_long 			xL, 
	A_long 			yL, 
	PF_PixelFloat 	*inP, 
	PF_PixelFloat 	*outP)
{
	register DistInfo	*siP		= (DistInfo*)refcon;
	PF_Err				err			= PF_Err_NONE;
	PF_Fixed			new_xFi 		= 0, 
						new_yFi 		= 0;
						
	AEGP_SuiteHandler suites(siP->in_data.pica_basicP);

	A_long cor_x = xL;
		A_long cor_y = siP->footage_height - yL;
		
		double fov_min_x = siP->_fov[0];
		double fov_max_x = siP->_fov[2];
		double fov_min_y = siP->_fov[1];
		double fov_max_y = siP->_fov[3];
		
		
		double x0,x1,y0,y1;
		
		x1 = x0 = (double(cor_x)+0.5 - fov_min_x) / double(fov_max_x - fov_min_x);
		y1 = y0 = (double(cor_y)+0.5 - fov_min_y) / double(fov_max_y - fov_min_y);
		
		
		if(siP->plugin)
		{
			switch (siP->distortion_mode) 
			{
				case DM_DISTORT:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
				case DM_UNDISTORT:
					siP->plugin->distort(x0, y0, x1, y1);
					break;
				default:
					siP->plugin->undistort(x0, y0, x1, y1);
					break;
			}
		}
		double xt = x1;
		double yt = y1;
		
		new_xFi = FLOAT2FIX(double(fov_max_x - fov_min_x)*xt - 0.5 + fov_min_x);
		new_yFi = FLOAT2FIX(double(siP->footage_height) - (double(fov_max_y - fov_min_y)*yt - 0.5 + fov_min_y));

	ERR(suites.SamplingFloatSuite1()->subpixel_sample_float(siP->in_data.effect_ref,
															new_xFi, 
															new_yFi, 
															&siP->samp_pb, 
															outP));
	return err;
}

static PF_Err 
Render (	
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output )
{
	PF_Err			err 		=	PF_Err_NONE;
	DistInfo		si;
	A_long			linesL		=	0;
	PF_EffectWorld	*inputP		=	&params[DIST_INPUT]->u.ld;
	
	PF_ParamDef _fbwidth,_fbheight,_fl,_lcox,_lcoy,_pa,dm;

	AEFX_CLR_STRUCT(si);
	
	si.footage_width = ((long)inputP->width << 16);
	si.footage_height = ((long)inputP->height << 16);
	si.ref = in_data->effect_ref;
	si.samp_pb.src = inputP;
	si.in_data = *in_data;
	
	tde4_ld_plugin* plug = 0;
	plug = (tde4_ld_plugin*) new tde4_ldp_radial_decentered_deg_4_cylindric<ldpk::vec2d,ldpk::mat2d>();
	si.plugin = plug;

	int nParams = 0;
	si.plugin->getNumParameters(nParams);
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FOCAL_LENGTH_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fl));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FB_WIDTH_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fbwidth));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FB_HEIGHT_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fbheight));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  LENS_CENTER_OFFSET_X_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_lcox));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  LENS_CENTER_OFFSET_Y_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_lcoy));
	
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  PIXEL_ASPECT_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_pa));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  DISTORT_MODE_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &dm));	
	
	si.distortion_mode = (DISTORTION_MODE)dm.u.pd.value;
	
	int parameter_offset = DISTORT_MODE_ID + 1;
	
	double v;
	v = FIX_2_FLOAT(_fl.u.fd.value)/10.0;
	si.plugin->setParameterValue("tde4_focal_length_cm",v);
	v = FIX_2_FLOAT(_fbwidth.u.fd.value);
	si.plugin->setParameterValue("tde4_filmback_width_cm",v);
	
	v = FIX_2_FLOAT(_fbheight.u.fd.value);
	si.plugin->setParameterValue("tde4_filmback_height_cm",v);
	
	v = FIX_2_FLOAT(_lcox.u.fd.value);
	si.plugin->setParameterValue("tde4_lens_center_offset_x_cm",v);
	
	v = FIX_2_FLOAT(_lcoy.u.fd.value);
	si.plugin->setParameterValue("tde4_lens_center_offset_y_cm",v);
	
	v = FIX_2_FLOAT(_pa.u.fd.value);
	si.plugin->setParameterValue("tde4_pixel_aspect",v);
	si.plugin->setParameterValue("tde4_custom_focus_distance_cm", 100);
	nParams = 0;
	si.plugin->getNumParameters(nParams);
	
	for(int i =0;i < nParams;i++)
	{
		char paramName[1000];
		si.plugin->getParameterName(i, paramName);
		tde4_ldp_ptype type;
		si.plugin->getParameterType(paramName, type);
		PF_ParamDef _value;
		ERR(PF_CHECKOUT_PARAM(	in_data, 
							  parameter_offset++, 
							  in_data->current_time, 
							  in_data->time_step, 
							  in_data->time_scale, 
							  &_value));
		
		switch (type) {
			case TDE4_LDP_STRING:
				break;
			case TDE4_LDP_DOUBLE:
			{
				double pDefault = FIX_2_FLOAT(_value.u.fd.value);
				si.plugin->setParameterValue(paramName,pDefault);
			}
				break;
			case TDE4_LDP_INT:
			{
				int pDefault;
				si.plugin->getParameterDefaultValue(paramName,pDefault);
				si.plugin->setParameterValue(paramName,pDefault);
			}
			case TDE4_LDP_FILE:
				break;
			case TDE4_LDP_TOGGLE:
				break;
			case TDE4_LDP_ADJUSTABLE_DOUBLE:
			{
				double pDefault = FIX_2_FLOAT(_value.u.fd.value);
				si.plugin->setParameterValue(paramName,pDefault);
				
				//std::cout << paramName << ":" << pDefault << std::endl;
			}
				break;
				
			default:
				break;
		}
		
	}
	si.plugin->initializeParameters();

		
	AEGP_SuiteHandler	suites(in_data->pica_basicP);		

	linesL = output->extent_hint.bottom - output->extent_hint.top;

	ERR(suites.Iterate8Suite1()->iterate(	in_data,
												0, 
												linesL, 
												inputP, 
												&output->extent_hint,
												(void*)&si, 
												DistortImage8, 
												output));

	delete si.plugin;
	si.plugin = 0;

	return err;
}

static PF_Err 
RespondtoAEGP ( 	
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output,
	void*			extraP)
{
	PF_Err			err = PF_Err_NONE;
	
	AEGP_SuiteHandler suites(in_data->pica_basicP);
	
	suites.ANSICallbacksSuite1()->sprintf(	out_data->return_msg, 
											"%s",	
											reinterpret_cast<A_char*>(extraP));

	return err;
}

static PF_Err
SmartRender(
	PF_InData				*in_data,
	PF_OutData				*out_data,
	PF_SmartRenderExtra		*extra)

{

	PF_Err				err		= PF_Err_NONE,
						err2	= PF_Err_NONE;
	PF_ParamDef _fbwidth,_fbheight,_fl,_lcox,_lcoy,_pa,dm,fov_min,fov_max;
						
	AEGP_SuiteHandler 	suites(in_data->pica_basicP);
	PF_EffectWorld		*input_worldP	= NULL, 
						*output_worldP  = NULL;
	PF_WorldSuite2		*wsP			= NULL;
	PF_PixelFormat		format			= PF_PixelFormat_INVALID;

	PF_Point			origin;

	DistInfo	*infoP = reinterpret_cast<DistInfo*>(suites.HandleSuite1()->host_lock_handle(reinterpret_cast<PF_Handle>(extra->input->pre_render_data)));

	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FOCAL_LENGTH_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fl));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FB_WIDTH_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fbwidth));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FB_HEIGHT_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_fbheight));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FOV_MIN_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &fov_min));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  FOV_MAX_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &fov_max));
						  
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  LENS_CENTER_OFFSET_X_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_lcox));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  LENS_CENTER_OFFSET_Y_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_lcoy));
	
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  PIXEL_ASPECT_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &_pa));
	ERR(PF_CHECKOUT_PARAM(	in_data, 
						  DISTORT_MODE_ID, 
						  in_data->current_time, 
						  in_data->time_step, 
						  in_data->time_scale, 
						  &dm));	
	
	infoP->distortion_mode = (DISTORTION_MODE)dm.u.pd.value;
	
	infoP->_fov[0] = FIX_2_FLOAT(fov_min.u.td.x_value);
	infoP->_fov[1] = FIX_2_FLOAT(fov_min.u.td.y_value);
	infoP->_fov[2] = FIX_2_FLOAT(fov_max.u.td.x_value);
	infoP->_fov[3] = FIX_2_FLOAT(fov_max.u.td.y_value);
	
	
	int parameter_offset = DISTORT_MODE_ID + 1;
	
	tde4_ld_plugin* plug = 0;
	
	plug = (tde4_ld_plugin*) new tde4_ldp_radial_decentered_deg_4_cylindric<ldpk::vec2d,ldpk::mat2d>();
	
	infoP->plugin = plug;

	double v;
	v = FIX_2_FLOAT(_fl.u.fd.value)/10.0;
	infoP->plugin->setParameterValue("tde4_focal_length_cm",v);
	v = FIX_2_FLOAT(_fbwidth.u.fd.value);
	infoP->plugin->setParameterValue("tde4_filmback_width_cm",v);
	
	v = FIX_2_FLOAT(_fbheight.u.fd.value);
	infoP->plugin->setParameterValue("tde4_filmback_height_cm",v);
	
	v = FIX_2_FLOAT(_lcox.u.fd.value);
	infoP->plugin->setParameterValue("tde4_lens_center_offset_x_cm",v);
	
	v = FIX_2_FLOAT(_lcoy.u.fd.value);
	infoP->plugin->setParameterValue("tde4_lens_center_offset_y_cm",v);
	
	v = FIX_2_FLOAT(_pa.u.fd.value);
	infoP->plugin->setParameterValue("tde4_pixel_aspect",v);
	infoP->plugin->setParameterValue("tde4_custom_focus_distance_cm", 100);
	int nParams = 0;
	infoP->plugin->getNumParameters(nParams);
	
	for(int i =0;i < nParams;i++)
	{
		char paramName[1000];
		infoP->plugin->getParameterName(i, paramName);
		tde4_ldp_ptype type;
		infoP->plugin->getParameterType(paramName, type);
		PF_ParamDef _value;
		ERR(PF_CHECKOUT_PARAM(	in_data, 
							  parameter_offset++, 
							  in_data->current_time, 
							  in_data->time_step, 
							  in_data->time_scale, 
							  &_value));
		
		switch (type) {
			case TDE4_LDP_STRING:
				break;
			case TDE4_LDP_DOUBLE:
			{
				double pDefault = FIX_2_FLOAT(_value.u.fd.value);
				infoP->plugin->setParameterValue(paramName,pDefault);
			}
				break;
			case TDE4_LDP_INT:
			{
				int pDefault;
				infoP->plugin->getParameterDefaultValue(paramName,pDefault);
				infoP->plugin->setParameterValue(paramName,pDefault);
			}
			case TDE4_LDP_FILE:
				break;
			case TDE4_LDP_TOGGLE:
				break;
			case TDE4_LDP_ADJUSTABLE_DOUBLE:
			{
				double pDefault = FIX_2_FLOAT(_value.u.fd.value);
				infoP->plugin->setParameterValue(paramName,pDefault);
				
				//std::cout << paramName << ":" << pDefault << std::endl;
			}
				break;
				
			default:
				break;
		}
		
	}
	
	infoP->plugin->initializeParameters();
	
	if (infoP){
		if (!infoP->no_opB){
			// checkout input & output buffers.
			ERR((extra->cb->checkout_layer_pixels(	in_data->effect_ref, DIST_INPUT, &input_worldP)));

			ERR(extra->cb->checkout_output(	in_data->effect_ref, &output_worldP));
			
			if (!err && output_worldP){
				ERR(AEFX_AcquireSuite(	in_data, 
										out_data, 
										kPFWorldSuite, 
										kPFWorldSuiteVersion2, 
										"Couldn't load suite.",
										(void**)&wsP));

				infoP->ref 			= in_data->effect_ref;
				infoP->samp_pb.src 	= input_worldP;
				infoP->in_data 		= *in_data;

				ERR(wsP->PF_GetPixelFormat(input_worldP, &format));

				origin.h = (A_short)(in_data->output_origin_x);
				origin.v = (A_short)(in_data->output_origin_y);

				if (!err){
					switch (format) {
					
						case PF_PixelFormat_ARGB128:
						
							ERR(suites.IterateFloatSuite1()->iterate_origin(in_data,
																			0,
																			output_worldP->height,
																			input_worldP,
																			&output_worldP->extent_hint,
																			&origin,
																			(void*)(infoP),
																			DistortImageFloat,
																			output_worldP));
							break;
							
						case PF_PixelFormat_ARGB64:
					
							ERR(suites.Iterate16Suite1()->iterate_origin(	in_data,
																			0,
																			output_worldP->height,
																			input_worldP,
																			&output_worldP->extent_hint,
																			&origin,
																			(void*)(infoP),
																			DistortImage16,
																			output_worldP));
							break;
							
						case PF_PixelFormat_ARGB32:
							 
							
							ERR(suites.Iterate8Suite1()->iterate_origin(	in_data,
																			0,
																			output_worldP->height,
																			input_worldP,
																			&output_worldP->extent_hint,
																			&origin,
																			(void*)(infoP),
																			DistortImage8,
																			output_worldP));
							break;

						default:
							err = PF_Err_BAD_CALLBACK_PARAM;
							break;
					}
				}

			}
		} else {
			// copy input buffer;
			ERR(PF_COPY(input_worldP, output_worldP, NULL, NULL));
		}
	} else {
		err = PF_Err_BAD_CALLBACK_PARAM;
	}
	ERR2(AEFX_ReleaseSuite(	in_data,
							out_data,
							kPFWorldSuite, 
							kPFWorldSuiteVersion2, 
							"Couldn't release suite."));
	
	delete infoP->plugin;
	infoP->plugin = 0;
	return err;
}

static PF_Err
PreRender(
	PF_InData				*in_data,
	PF_OutData				*out_data,
		  PF_ParamDef		*params[],
	PF_PreRenderExtra		*extra)
{
	PF_Err err = PF_Err_NONE;
	PF_RenderRequest req = extra->input->output_request;
	PF_CheckoutResult in_result;

	AEGP_SuiteHandler suites(in_data->pica_basicP);

	PF_Handle	infoH	=	suites.HandleSuite1()->host_new_handle(sizeof(DistInfo));
	
	if (infoH){
		
		DistInfo	*infoP = reinterpret_cast<DistInfo*>(suites.HandleSuite1()->host_lock_handle(infoH));
		
		if (infoP){
			extra->output->pre_render_data = infoH;


			if (!err){
				req.preserve_rgb_of_zero_alpha = TRUE;	//	Hey, we care.

				ERR(extra->cb->checkout_layer(	in_data->effect_ref,
												DIST_INPUT,
												DIST_INPUT,
												&req,
												in_data->current_time,
												in_data->time_step,
												in_data->time_scale,
												&in_result));

				if (!err){
					AEFX_CLR_STRUCT(*infoP);

					PF_Fixed 	widthF	= ABS(in_result.max_result_rect.right - in_result.max_result_rect.left),
								heightF = ABS(in_result.max_result_rect.bottom - in_result.max_result_rect.top);

					infoP->footage_width 		= (widthF ) ;

					infoP->footage_height 		= (heightF ) ;

					UnionLRect(&in_result.result_rect, 		&extra->output->result_rect);
					UnionLRect(&in_result.max_result_rect, 	&extra->output->max_result_rect);	

				}
			}
		} else {
			err = PF_Err_OUT_OF_MEMORY;
		}		
	} else {
		err = PF_Err_OUT_OF_MEMORY;
	}
	return err;
	
}



DllExport	PF_Err 
EntryPointFunc (
	PF_Cmd			cmd,
	PF_InData		*in_data,
	PF_OutData		*out_data,
	PF_ParamDef		*params[],
	PF_LayerDef		*output,
	void			*extraP)
{
	PF_Err		err = PF_Err_NONE;
	
	try
	{
		switch (cmd) {
			case PF_Cmd_ABOUT:
				err = About(in_data, 
							out_data, 
							params, 
							output);
				break;
			case PF_Cmd_GLOBAL_SETUP:
				err = GlobalSetup(	in_data, 
									out_data,
									params, 
									output);
				break;
			case PF_Cmd_PARAMS_SETUP:
				err = ParamsSetup(	in_data, 
									out_data, 
									params, 
									output);
				break;
			case PF_Cmd_RENDER:
				err = Render(	in_data, 
								out_data, 
								params, 
								output);
				break;
			case PF_Cmd_FRAME_SETUP:
				err = FrameSetup(	in_data, 
								out_data, 
								params, 
								output);
				break;
			case PF_Cmd_SMART_PRE_RENDER:
				err = PreRender(in_data, 
								out_data,
								params,
								reinterpret_cast<PF_PreRenderExtra*>(extraP));
				break;
				
			case PF_Cmd_SMART_RENDER:
				err = SmartRender(	in_data, 
									out_data, 
									reinterpret_cast<PF_SmartRenderExtra*>(extraP));
				break;

				
			case PF_Cmd_COMPLETELY_GENERAL:
				err = RespondtoAEGP(in_data, 
									out_data, 
									params, 
									output, 
									extraP);
		}
	}
	catch(PF_Err &thrown_err)
	{
		err = thrown_err;
	}
	return err;
}



